#!/usr/bin/perl -w

# ICC::Profile::ncl2 test module / 2010-08-07
#
# Copyright © 2004-2020 by William B. Birkett

use strict;

use Test::More tests => 1;

# does module load
BEGIN { use_ok('ICC::Profile::ncl2') };

##### more tests needed #####

