#!/usr/bin/perl -w

# ICC::Profile::spline test module / 2018-08-31
#
# Copyright © 2004-2020 by William B. Birkett

use strict;

use File::Spec;
use Test::More tests => 3;

# local variables
my ($tag);

# does module load
BEGIN { use_ok('ICC::Support::spline') };

# test class methods
can_ok('ICC::Support::spline', qw(new inv write_fh size transform inverse derivative parametric header input output monotonic table curv normalize update sdump));

# make empty spline object
$tag = ICC::Support::spline->new;

# test object class
isa_ok($tag, 'ICC::Support::spline');

##### more tests needed ######


