#!/usr/bin/perl -w

# ICC::Profile::bern test module / 2018-08-11
#
# Copyright © 2004-2020 by William B. Birkett

use strict;

use File::Spec;
use Test::More tests => 3;

# local variables
my ($tag);

# does module load
BEGIN { use_ok('ICC::Support::bern') };

# test class methods
can_ok('ICC::Support::bern', qw(new inv write_fh size transform inverse derivative parametric header input output table normalize monotonic update curv sdump));

# make empty bern object
$tag = ICC::Support::bern->new;

# test object class
isa_ok($tag, 'ICC::Support::bern');

##### more tests needed ######


